#include "util.h"

#include <arpa/inet.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <netdb.h>
#include <stdlib.h>
#include <stdlib.h>
#include <string.h>
#include <sys/time.h>
#include <syslog.h>
#include <unistd.h>

#include "log.h"

#include <digitalSTROM/ds485.h>

/* TODO function to drop root privileges */

int sprintf_dsuid(char *buf, const dsuid_t *id)
{
    char *p = buf;
    unsigned n;

    for (n = 0; n < sizeof(dsuid_t); n++) {
        p += sprintf(p, "%02x", id->id[n]);
    }
    return (p - buf);
}

static char format[] = "%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx%02hhx";
int scanf_dsuid(const char *buf, dsuid_t *dsuid)
{
    int ret;
    ret = sscanf(buf, format,
           &dsuid->id[0],
           &dsuid->id[1],
           &dsuid->id[2],
           &dsuid->id[3],
           &dsuid->id[4],
           &dsuid->id[5],
           &dsuid->id[6],
           &dsuid->id[7],
           &dsuid->id[8],
           &dsuid->id[9],
           &dsuid->id[10],
           &dsuid->id[11],
           &dsuid->id[12],
           &dsuid->id[13],
           &dsuid->id[14],
           &dsuid->id[15],
           &dsuid->id[16]);

    if (ret != sizeof(dsuid->id)) {
        return -EINVAL;
    }
    return 0;
}

int sprintf_container(char *buf, const ds485_container_t *container)
{
    int n;
    char *p;

    p = buf;
    p += sprintf_dsuid(p, &container->sourceId);
    p += sprintf(p, " -> ");
    p += sprintf_dsuid(p, &container->destinationId);
    p += sprintf(p, " [len %02X, tid %02X] : ", container->length, container->transactionId);
    for (n = 0; n < 16 && n < container->length; n++)
        p += sprintf(p, "%02x", container->data[n]);
    *p = 0;

    return (p - buf);
}

int sprintf_netlib_packet(char *buf, const ds485n_packet_t *packet)
{
    int n;
    char *p;

    p = buf;
    p += sprintf_dsuid(p, &packet->sourceId);
    p += sprintf(p, " -> ");
    p += sprintf_dsuid(p, &packet->destinationId);
    p += sprintf(p, " [len %02X, tid %02X] : ", packet->length, packet->transactionId);
    for (n = 0; n < 16 && n < packet->length; n++)
        p += sprintf(p, "%02x", packet->data[n]);
    *p = 0;

    return (p - buf);
}

double timestamp(void)
{
    struct timeval tv;

    gettimeofday(&tv, NULL);
    return (tv.tv_sec + tv.tv_usec*1e-6);
}

int snprintf_peername(char *buf, int len, int fd)
{
    struct sockaddr_in fsin;
    socklen_t alen = (socklen_t)sizeof(fsin);
    int r;

    r = getpeername(fd, (struct sockaddr *) &fsin, &alen);
    if (r < 0) {
        report(LOG_INFO, "getpeername() = %d, error = %s (%d)\n", r, strerror(errno), errno);
        return errno;
    }

    r = snprintf(buf, len, "%s:%d", inet_ntoa(fsin.sin_addr),
                 ntohs(fsin.sin_port));
    if (r < 0) {
        report(LOG_NOTICE, "snprintf failed %s\n", strerror(errno));
        return r;
    } else if (r >= len) {
        report(LOG_WARNING, "%s buffer too small %d > %d\n", r, len);
        buf[len - 1] = '0';
        return 0; /* ignore */
    }

    return 0;
}

int daemonize(void)
{
    int fd;
    pid_t pid;

    switch (pid = fork()) {
    case -1:
        return -1;
    case 0:        /* child side */
        break;
    default:    /* parent side */
        exit(0);
    }

    if (setsid() == -1)
        return -1;
    if (chdir("/") == -1)
        return -1;

    if ((fd = open("/dev/null", O_RDWR, 0)) != -1) {
        dup2(fd, STDIN_FILENO);
        dup2(fd, STDOUT_FILENO);
        dup2(fd, STDERR_FILENO);
        if (fd > 2)
            close(fd);
    }
    return 0;
}

int sscanf_port(const char* str_port)
{
    long val;
    char *endptr;

    errno = 0;    /* To distinguish success/failure after call */
    val = strtol(str_port, &endptr, 10);

    /* Check for various possible errors */
    if ((errno == ERANGE && (val == LONG_MAX || val == LONG_MIN))
        || (errno != 0 && val == 0)) {
        report(LOG_WARNING, "invalid port value %s\n", val);
        return -ERANGE;
    }

    if (endptr == str_port) {
        report(LOG_NOTICE, "No digits were found\n");
        return -EINVAL;
    }

    if (val <= 0 || val > USHRT_MAX) {
        report(LOG_WARNING, "port out of range %s\n", val);
        return -ERANGE;
    }

    return val;
}

int service_to_port(const char *service, const char *protocol)
{
    struct servent *pse;
    int port;

    if ((pse = getservbyname(service, protocol)) == NULL) {
        report(LOG_INFO, "Can't get \"%s\" service entry\n", service);
        return 0;
    }

    port = htons(ntohs((in_port_t)pse->s_port));
    return port;
}

int passivesock(int port, char *protocol, int qlen, int listenGlobal)
{
    struct protoent *ppe;
    struct sockaddr_in sin;
    int s, type, proto, one = 1;

    memset((char *) &sin, 0, sizeof(sin));
    sin.sin_family = (sa_family_t)AF_INET;

    if (listenGlobal)
        sin.sin_addr.s_addr = htonl(INADDR_ANY);
    else
        sin.sin_addr.s_addr = htonl(INADDR_LOOPBACK);

    sin.sin_port = htons(port);

    ppe = getprotobyname(protocol);
    if (strcmp(protocol, "udp") == 0) {
        type = SOCK_DGRAM;
        proto = (ppe) ? ppe->p_proto : IPPROTO_UDP;
    } else {
        type = SOCK_STREAM;
        proto = (ppe) ? ppe->p_proto : IPPROTO_TCP;
    }
    if ((s = socket(PF_INET, type, proto)) < 0) {
        report(LOG_ERR, "Can't create socket\n");
        return -1; /* TODO */
    }
    if (setsockopt(s,SOL_SOCKET,SO_REUSEADDR,(char *)&one,(int)sizeof(one)) == -1) {
        report(LOG_ERR, "Error: SETSOCKOPT SO_REUSEADDR\n");
        return -1; /* TODO */
    }
    if (bind(s, (struct sockaddr *) &sin, (int)sizeof(sin)) < 0) {
        report(LOG_ERR, "Can't bind to port %d\n", port);
        if (errno == EADDRINUSE) {
            report(LOG_ERR, "Maybe daemon already running!\n");
        }
        return -1; /* TODO */
    }
    if (type == SOCK_STREAM && listen(s, qlen) < 0) {
        report(LOG_ERR, "Can't listen on port %d\n", port);
        return -1; /* TODO */
    }

    int opts = fcntl(s, F_GETFL);
    if (opts > 0) {
        fcntl(s, F_SETFL, opts | O_NONBLOCK);
    }
    #ifdef IPTOS_LOWDELAY
    {
        int opt = IPTOS_LOWDELAY;
        setsockopt(s, IPPROTO_IP, IP_TOS, &opt, sizeof opt);
    }
    #endif
    #ifdef TCP_NODELAY
    {
        if (type == SOCK_STREAM)
            setsockopt(s, IPPROTO_TCP, TCP_NODELAY, &one, sizeof one);
    }
    #endif

    return s;
}

#if 0
int main()
{
    DSID_STR_BUF(buf);
    dsid_t dsid;
    scanf_dsid("123456789012345678901234", &dsid);
    sprintf_dsid(buf, &dsid);
    printf("dsid: %s\n", buf);

}
#endif
